package ru.dbotthepony.mc.otm.block.entity.blackhole

import net.minecraft.client.Minecraft
import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.network.Connection
import net.minecraft.network.protocol.PacketFlow
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.item.ItemEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.Explosion
import net.minecraft.world.level.ExplosionDamageCalculator
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.levelgen.structure.BoundingBox
import net.minecraft.world.level.material.FluidState
import net.minecraft.world.level.saveddata.SavedData
import net.minecraft.world.phys.AABB
import net.minecraft.world.phys.Vec3
import net.minecraftforge.event.TickEvent
import net.minecraftforge.eventbus.api.SubscribeEvent
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlackHoleExplosionQueue.Companion.factory
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.matter.MatterRegistry
import ru.dbotthepony.mc.otm.set
import java.util.*
import kotlin.collections.ArrayList
import kotlin.math.cos
import kotlin.math.pow
import kotlin.math.sin
import kotlin.math.sqrt

private object BlackHoleExplosionDamageCalculator : ExplosionDamageCalculator() {
	override fun getBlockExplosionResistance(
		explosion: Explosion,
		getter: BlockGetter,
		pos: BlockPos,
		state: BlockState,
		fstate: FluidState
	): Optional<Float> {
		return if (state.isAir && fstate.isEmpty) Optional.empty() else Optional.of(
			Math.sqrt(
				Math.max(
					0f,
					Math.max(
						state.getExplosionResistance(getter, pos, explosion),
						fstate.getExplosionResistance(getter, pos, explosion)
					)
				).toDouble()
			).toFloat()
		)
	}
}

private data class RingExplosion(val x: Double, val y: Double, val z: Double, val radius: Double, val strength: Float) {
	fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["x"] = x
			it["y"] = y
			it["z"] = z
			it["radius"] = radius
			it["strength"] = strength
		}
	}

	fun explode(queue: BlackHoleExplosionQueue) {
		val fragments = radius.toInt() * 8
		val stackStep = Math.PI / fragments.toDouble()
		val sectorStep = Math.PI / fragments.toDouble() * 2

		for (stack in 0 until fragments) {
			val stackAngle = Math.PI / 2 - stack * stackStep
			val xy = radius * 15 * cos(stackAngle)
			val z = radius * 15 * sin(stackAngle)

			for (sector in 0 until fragments) {
				val sectorAngle = sector * sectorStep

				queue.explode(
					this.x + xy * cos(sectorAngle),
					this.y + xy * sin(sectorAngle),
					this.z + z,
					strength
				)
			}
		}
	}

	companion object {
		@JvmStatic
		fun deserializeNBT(tag: CompoundTag): RingExplosion {
			return RingExplosion(
				tag.getDouble("x"),
				tag.getDouble("y"),
				tag.getDouble("z"),
				tag.getDouble("radius"),
				tag.getFloat("strength")
			)
		}
	}
}

private data class QueuedExplosion(val x: Double, val y: Double, val z: Double, val radius: Float) {
	fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["x"] = x
			it["y"] = y
			it["z"] = z
			it["radius"] = radius
		}
	}

	fun explode(level: Level) {
		level.explode(
			null,
			Registry.DAMAGE_HAWKING_RADIATION,
			BlackHoleExplosionDamageCalculator,
			x,
			y,
			z,
			radius,
			false,
			Explosion.BlockInteraction.DESTROY
		)
	}

	companion object {
		fun deserializeNBT(tag: CompoundTag): QueuedExplosion {
			return QueuedExplosion(tag.getDouble("x"), tag.getDouble("y"), tag.getDouble("z"), tag.getFloat("radius"))
		}
	}
}

class BlackHoleExplosionQueue(private val level: ServerLevel) : SavedData() {
	private var indexExplosion = 0
	private var indexRing = 0
	private val explosions = ArrayList<QueuedExplosion>()
	private val rings = ArrayList<RingExplosion>()

	override fun save(tag: CompoundTag): CompoundTag {
		val listExplosions = ListTag()
		val listRings = ListTag()

		for (i in indexExplosion until explosions.size)
			listExplosions.add(explosions[i].serializeNBT())

		for (i in indexRing until rings.size)
			listRings.add(rings[i].serializeNBT())

		tag["explosions"] = listExplosions
		tag["rings"] = listRings

		return tag
	}

	fun load(tag: CompoundTag) {
		explosions.clear()
		rings.clear()

		indexExplosion = 0
		indexRing = 0

		for (explosion in tag.getList("explosions", Tag.TAG_COMPOUND.toInt()))
			explosions.add(QueuedExplosion.deserializeNBT(explosion as CompoundTag))

		for (ring in tag.getList("rings", Tag.TAG_COMPOUND.toInt()))
			rings.add(RingExplosion.deserializeNBT(ring as CompoundTag))
	}

	fun explode(x: Double, y: Double, z: Double, radius: Float) {
		explosions.add(QueuedExplosion(x, y, z, radius))
		isDirty = true
	}

	fun explodeRing(x: Double, y: Double, z: Double, radius: Double, strength: Float) {
		rings.add(RingExplosion(x, y, z, radius, strength))
		isDirty = true
	}

	fun tick() {
		if (explosions.size != 0) {
			setDirty()
			var iterations = 0

			for (i in indexExplosion until explosions.size) {
				explosions[i].explode(level)
				indexExplosion++

				if (iterations++ == 4) {
					break
				}
			}

			if (indexExplosion >= explosions.size) {
				indexExplosion = 0
				explosions.clear()
			}
		} else if (rings.size != 0) {
			if (indexRing >= rings.size) {
				indexRing = 0
				rings.clear()
			} else {
				rings[indexRing++].explode(this)
			}
		}
	}

	companion object {
		@JvmStatic
		fun factory(level: ServerLevel): BlackHoleExplosionQueue {
			return level.dataStorage.computeIfAbsent(
				{
					val factory = BlackHoleExplosionQueue(level)
					factory.load(it)
					factory
				},
				{ BlackHoleExplosionQueue(level) },
				"otm_blackhole_explosion_queue"
			)
		}

		@SubscribeEvent
		fun onWorldTick(event: TickEvent.WorldTickEvent) {
			if (event.phase == TickEvent.Phase.START && event.world is ServerLevel) {
				factory(event.world as ServerLevel).tick()
			}
		}
	}
}

class BlockEntityBlackHole(p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(Registry.BlockEntities.BLACK_HOLE, p_155229_, p_155230_) {
	var mass = BASELINE_MASS
		set(mass) {
			if (mass <= Fraction.ZERO) {
				collapse()
				return
			}

			field = mass

			setChanged()
			gravitationStrength = mass.div(BASELINE_MASS).toDouble()

			if (gravitationStrength > 1) {
				gravitationStrength = 20.0.coerceAtMost(sqrt(gravitationStrength))
			} else {
				gravitationStrength = 0.08.coerceAtLeast(gravitationStrength.pow(2.0))
			}

			val level = level

			if (level != null && !level.isClientSide && !suppress_updates)
				level.sendBlockUpdated(blockPos, blockState, blockState, Block.UPDATE_CLIENTS)

			affectedBounds = BoundingBox(
				(-30 * gravitationStrength).toInt(),
				(-30 * gravitationStrength).toInt(),
				(-30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt()
			).move(
				blockPos
			)

			affectedBoundsAABB = AABB.of(affectedBounds)
		}

	var gravitationStrength = 1.0
		private set

	private var suppress_updates = true
	var spin_direction = false

	fun collapse() {
		val level = level as? ServerLevel ?: return

		level.setBlock(blockPos, Blocks.AIR.defaultBlockState(), Block.UPDATE_ALL)

		if (gravitationStrength > 0.25) {
			val x0 = blockPos.x + 0.5
			val y0 = blockPos.y + 0.5
			val z0 = blockPos.z + 0.5
			val queue = factory(level)
			var radius = 0

			while (radius < Math.ceil(gravitationStrength * 4)) {
				queue.explodeRing(
					x0,
					y0,
					z0,
					radius.toDouble(), Math.min(20.0, Math.max(1.0, (gravitationStrength * 4 - radius) * 20)).toFloat()
				)

				radius++
			}
		} else {
			level.explode(
				null,
				Registry.DAMAGE_HAWKING_RADIATION,
				null,
				blockPos.x + 0.5,
				blockPos.y + 0.5,
				blockPos.z + 0.5,
				gravitationStrength.toFloat() * 60,
				false,
				Explosion.BlockInteraction.DESTROY
			)
		}
	}

	// shared functions
	fun writeBlackHoleData(tag: CompoundTag): CompoundTag {
		tag["mass"] = mass.serializeNBT()
		tag["spin_direction"] = spin_direction
		return tag
	}

	fun readBlackHoleData(tag: CompoundTag) {
		tag["mass"]?.let {
			mass = Fraction.deserializeNBT(it)
		}

		spin_direction = tag.getBoolean("spin_direction")
	}

	// disk io
	public override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		writeBlackHoleData(nbt)
	}

	override fun load(tag: CompoundTag) {
		super.load(tag)
		readBlackHoleData(tag)
	}

	// received either by game engine (from getUpdateTag on ClientLevelPacket)
	// or by onDataPacket
	override fun handleUpdateTag(tag: CompoundTag) {
		super.handleUpdateTag(tag)
		readBlackHoleData(tag)
	}

	// called by game engine for ClientLevelPacket
	override fun getUpdateTag(): CompoundTag {
		return writeBlackHoleData(super.getUpdateTag())
	}

	// called by game engine on block updates
	override fun getUpdatePacket(): ClientboundBlockEntityDataPacket? {
		return ClientboundBlockEntityDataPacket.create(this)
	}

	// called by game engine by forge patches when ClientboundBlockEntityDataPacket is received
	override fun onDataPacket(net: Connection, pkt: ClientboundBlockEntityDataPacket) {
		super.onDataPacket(net, pkt)

		if (net.receiving == PacketFlow.CLIENTBOUND) {
			handleUpdateTag(pkt.tag!!)
		}
	}

	var affectedBounds = BoundingBox(0, 0, 0, 1, 1, 1)
		private set
	var affectedBoundsAABB = AABB.of(affectedBounds)
		private set

	private fun setDeltaMovement(living: Entity, center: Vec3, distance: Double, weaker: Boolean) {
		//final double mult = Math.min(2, (30 * this.gravitation_strength) / Math.max(1, Math.pow(distance, 2)));
		// Сила притяжения
		val mult = Math.pow(1 - distance / (30 * gravitationStrength), 2.0) * gravitationStrength / 8

		// Притяжение к ядру
		val delta = living.position().vectorTo(center).normalize()

		if (distance < gravitationStrength) {
			living.deltaMovement = living.deltaMovement.add(delta.multiply(mult, mult, mult))
		} else {
			// Закручивание
			val rotate = if (spin_direction) delta.yRot((Math.PI / 2).toFloat()) else delta.yRot((-Math.PI / 2).toFloat())

			if (weaker)
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.4f, mult * 0.4f, mult * 0.4f))
					.add(delta.multiply(mult * 0.33f, mult * 0.33f, mult * 0.33f))
			else
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.4f, mult * 0.4f, mult * 0.4f))
					.add(delta.multiply(mult, mult, mult))
		}
	}

	fun clientTick() {
		val ply = Minecraft.getInstance().player!!
		val center = Vec3.atCenterOf(blockPos)

		if (!ply.abilities.mayfly) {
			val distance = ply.position().distanceTo(center)
			setDeltaMovement(ply, center, distance, true)
		}

		for (item in level!!.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
			val distance = item.position().distanceTo(center)
			setDeltaMovement(item, center, distance, false)
		}
	}

	fun tick() {
		val level = level as ServerLevel
		suppress_updates = false

		val center = Vec3.atCenterOf(blockPos)

		for (living in level.getEntitiesOfClass(LivingEntity::class.java, affectedBoundsAABB)) {
			val distance = living.position().distanceTo(center)

			if (living !is Player || !living.abilities.mayfly) {
				setDeltaMovement(living, center, distance, living !is Player)
			}

			if (distance < gravitationStrength + 1) {
				living.hurt(Registry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat())
			}
		}

		for (item in level.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
			val distance = item.position().distanceTo(center)
			setDeltaMovement(item, center, distance, false)

			if (distance < gravitationStrength + 1) {
				if (item.hurt(Registry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat()) && item.isRemoved) {
					if (item.item.item === Registry.Items.GRAVITATIONAL_DISRUPTOR) {
						collapse()
					} else {
						val mass = MatterRegistry.getMatterValue(item.item)

						if (mass > Fraction.ZERO)
							this.mass += mass
					}
				}
			}
		}

		// шанс 1% что черная дыра потеряет 0.1 MtU каждую секунду * силу гравитации дыры ^ -1
		if (level.random.nextDouble() < 0.01 * 0.05 * (1 / gravitationStrength)) {
			this.mass += HAWKING_MASS_LOSE_STEP
		}
	}

	init {
		mass = mass
	}

	companion object {
		val BASELINE_MASS = Fraction(1_000)
		val HAWKING_MASS_LOSE_STEP = Fraction("-0.1")
	}
}